# Tests for Redock.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: August 6, 2013
# URL: https://github.com/xolox/python-redock

# Standard library modules.
import logging
import pipes
import subprocess
import unittest

# External dependencies.
import coloredlogs

# Modules included in our package.
from redock.api import Container, Image
from redock.logger import root_logger

class RedockTestCase(unittest.TestCase):

    def setUp(self):
        coloredlogs.install()
        root_logger.setLevel(logging.DEBUG)

    def test_image_coercion(self):
        image = Image.coerce('redock:test')
        self.assertEqual(image.repository, 'redock')
        self.assertEqual(image.tag, 'test')
        self.assertEqual(image.key, ('redock', 'test'))
        self.assertEqual(image.name, 'redock:test')
        self.assertEqual(image.unique_name, 'redock:test')

    def test_start_container(self):
        hostname = 'whatever'
        # Start a test container.
        container = Container('redock:test', hostname=hostname)
        container.start()
        try:
            # Try to connect to the container over SSH.
            ssh_client = subprocess.Popen(['ssh', container.ssh_alias, 'hostname'],
                                          stdin=subprocess.PIPE,
                                          stdout=subprocess.PIPE)
            stdout, stderr = ssh_client.communicate()
            # Check the results.
            self.assertEqual(ssh_client.returncode, 0)
            self.assertEqual(stdout.strip(), hostname)
        finally:
            # Kill the container.
            container.kill()
        # Verify that the container is no longer reachable over SSH.
        ssh_client = subprocess.Popen(['ssh', '-q', container.ssh_alias, 'hostname'],
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE)
        stdout, stderr = ssh_client.communicate()
        self.assertNotEqual(ssh_client.returncode, 0)

    def test_commit_and_delete_container(self):
        pathname = '/root/peter-was-here'
        # Start a test container.
        container = Container('redock:test')
        container.start()
        try:
            # Create a file on the container's file system.
            ssh_client = subprocess.Popen(['ssh', container.ssh_alias, 'touch %s' % pipes.quote(pathname)])
            self.assertEqual(ssh_client.wait(), 0)
            # Commit the container's file system to an image.
            container.commit()
            # Restart the container.
            container.kill()
            container.start()
            # Make sure the file still exists.
            ssh_client = subprocess.Popen(['ssh', container.ssh_alias, 'test -f %s' % pipes.quote(pathname)])
            self.assertEqual(ssh_client.wait(), 0)
        finally:
            # Kill the container.
            container.kill()
            # Delete the image.
            container.delete()

if __name__ == '__main__':
    unittest.main()

# vim: ts=4 sw=4 et



# Utility functions for Redock.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2013
# URL: https://github.com/xolox/python-redock

# Standard library modules.
import fcntl
import os.path
import pickle
import pipes
import re
import socket
import subprocess
import sys
import urllib

# External dependencies.
import netifaces
from humanfriendly import format_path

# Modules included in our package.
from redock.logger import get_logger

logger = get_logger(__name__)

# Directory on the host system with files generated by Redock.
REDOCK_CONFIG_DIR = os.path.expanduser('~/.redock')

# The absolute pathname of the serialized runtime configuration.
CONFIG_FILE = os.path.join(REDOCK_CONFIG_DIR, 'state.pickle')

# The version number of the runtime configuration format.
CONFIG_VERSION = 1

# The absolute pathname of the text file containing the selected Ubuntu mirror.
UBUNTU_MIRROR_FILE = os.path.join(REDOCK_CONFIG_DIR, 'ubuntu-mirror.txt')

# The absolute pathname of SSH public key generated by Redock.
PUBLIC_SSH_KEY = os.path.join(REDOCK_CONFIG_DIR, 'id_rsa.pub')

# The absolute pathname of SSH private key generated by Redock.
PRIVATE_SSH_KEY = os.path.join(REDOCK_CONFIG_DIR, 'id_rsa')

class Config(object):

    # TODO The Config class as it is written now is a bit tricky. I feel
    #      like it should be a lot simpler! I hope this is possible :-)

    """
    :py:class:`Config` encapsulates the bits of runtime configuration that
    Redock needs to persist to disk (to share state in between runs of Redock).
    UNIX file locking is used to guarantee that the datafile is not written to
    simultaneously by multiple processes (that could corrupt the state).

    To use this class to update the configuration, use it like a context
    manager, like this:

    >>> config = Config()
    >>> with config as state:
    ...   state['containers'].clear()

    When used like this, ``state`` is a dictionary which is saved to disk when
    the ``with`` block ends without raising an exception.
    """

    def __init__(self):
        self.logger = logger
        self.handle = None
        self.state = {}

    def load(self, exists=True):
        """
        Load the runtime configuration from disk. If the file doesn't exist yet
        an empty configuration is returned. The configuration contains a
        version number which enables graceful upgrades to the format.

        :returns: A dictionary with runtime configuration data.
        """
        if exists:
            self.logger.verbose("Loading runtime configuration from %s ..", format_path(CONFIG_FILE))
        state = {}
        handle = self.handle
        close = False
        if (not handle) and os.path.isfile(CONFIG_FILE):
            handle = open(CONFIG_FILE)
            close = True
        if exists and handle:
            state = pickle.load(handle)
            if close:
                handle.close()
        version = state.get('version', 0)
        if version == 0:
            state['containers'] = dict()
            state['version'] = CONFIG_VERSION
        self.logger.debug("Initialized configuration: %r", state)
        return state

    def __enter__(self):
        exists = os.path.isfile(CONFIG_FILE)
        self.handle = open(CONFIG_FILE, 'r+' if exists else 'w')
        fcntl.flock(self.handle.fileno(), fcntl.LOCK_EX)
        self.state = self.load(exists=exists)
        return self.state

    def __exit__(self, type, value, traceback):
        if type is None:
            self.logger.verbose("Saving configuration to %s ..", format_path(CONFIG_FILE))
            self.handle.seek(0)
            pickle.dump(self.state, self.handle)
            self.handle.truncate()
        else:
            self.logger.warn("Not saving configuration! (an exception was raised: %s)", value)
        fcntl.flock(self.handle.fileno(), fcntl.LOCK_UN)
        self.handle.close()
        self.handle = None

class RemoteTerminal(object):

    """
    Attach to a running Docker container and show the output of the command(s)
    inside the container on the host's terminal. Can be used as a context
    manager or by manually calling :py:func:`RemoteTerminal.attach()` and
    :py:func:`RemoteTerminal.detach()`.
    """

    def __init__(self, container_id):
        """
        Initialize the context manager for the ``docker attach`` process.

        :param container_id: The id of the container to attach to (a string).
        """
        self.container_id = container_id

    def attach(self):
        """
        Start the ``docker attach`` subprocess.
        """
        logger.verbose("Attaching to terminal of container %s ..", summarize_id(self.container_id))
        self.subprocess = subprocess.Popen(['docker', 'attach', self.container_id], stdin=open(os.devnull), stdout=sys.stderr)

    def detach(self):
        """
        Kill the ``docker attach`` subprocess.
        """
        logger.verbose("Detaching from container %s ..", summarize_id(self.container_id))
        self.subprocess.kill()

    def __enter__(self):
        self.attach()

    def __exit__(self, type, value, traceback):
        self.detach()

def select_ubuntu_mirror(force=False):
    """
    Find an Ubuntu mirror that is geographically close to the current
    location for use inside Docker containers. We remember the choice in a
    file on the host system so that we always configure the same mirror in
    Docker containers (if you change the mirror, ``apt-get`` has to
    download all package metadata again, wasting a lot of time).
    """
    create_configuration_directory()
    if not os.path.isfile(UBUNTU_MIRROR_FILE):
        url = 'http://mirrors.ubuntu.com/mirrors.txt'
        logger.debug("Finding nearby Ubuntu package mirror using %s ..", url)
        mirror = urllib.urlopen(url).readline().strip()
        with open(UBUNTU_MIRROR_FILE, 'w') as handle:
            handle.write('%s\n' % mirror)
    with open(UBUNTU_MIRROR_FILE) as handle:
        mirror = handle.read().strip()
        logger.debug("Selected Ubuntu package mirror: %s", mirror)
    return mirror

def get_ssh_public_key():
    """
    Get the contents of the SSH public key generated by Redock for use inside
    containers. If the SSH key pair hasn't been generated yet, it will be
    generated using :py:func:`generate_ssh_key_pair()`.

    :returns: The contents of the ``id_rsa.pub`` file.
    """
    if not os.path.isfile(PUBLIC_SSH_KEY):
        generate_ssh_key_pair()
    with open(PUBLIC_SSH_KEY) as handle:
        return handle.read().strip()

def generate_ssh_key_pair():
    """
    Generate an SSH key pair for communication between the host system and
    containers created with Redock. Requires the ``ssh-keygen`` program.
    """
    create_configuration_directory()
    logger.verbose("Checking if we need to generate a new SSH key pair ..")
    if os.path.isfile(PRIVATE_SSH_KEY):
        logger.verbose("SSH key pair was previously generated: %s", format_path(PRIVATE_SSH_KEY))
        return
    logger.info("No existing SSH key pair found, generating new key: %s", format_path(PRIVATE_SSH_KEY))
    command = ['ssh-keygen', '-t', 'rsa', '-f', PRIVATE_SSH_KEY, '-N', '', '-C', 'root@%s' % socket.gethostname()]
    ssh_keygen = subprocess.Popen(command)
    if ssh_keygen.wait() != 0:
        msg = "Failed to generate SSH key pair! (command exited with code %d: %s)"
        raise Exception, msg % (ssh_keygen.returncode, quote_command_line(command))

def find_local_ip_addresses():
    """
    To connect to a running Docker container over TCP we need to connect to a
    specific port number on an IP address associated with a local network
    interface on the host system (specifically *not* a loop back interface).

    :returns: A :py:class:`set` of IP addresses associated with local network
              interfaces.
    """
    ip_addresses = set()
    for name in sorted(netifaces.interfaces(), key=str.lower):
        for addresses in netifaces.ifaddresses(name).values():
            for properties in addresses:
                address = properties.get('addr')
                # As mentioned above we're specifically *not* interested in loop back interfaces.
                if address.startswith('127.'):
                    continue
                # I'm not interested in IPv6 addresses right now.
                if ':' in address:
                    continue
                if address:
                    ip_addresses.add(address)
    return ip_addresses

def apt_get_install(*packages):
    """
    Generate a command to install the given packages with ``apt-get``.

    :param packages: The names of the package(s) to be installed.
    :returns: The ``ap-get`` command line as a single string.
    """
    command = ['DEBIAN_FRONTEND=noninteractive',
               'apt-get', 'install', '-q', '-y',
               '--no-install-recommends']
    return quote_command_line(command + list(packages))

def quote_command_line(command):
    """
    Quote the tokens in a shell command line.

    :param command: A list with the command name and arguments.
    :returns: The command line as a single string.
    """
    return ' '.join(pipes.quote(s) for s in command)

def summarize_id(id):
    """
    Docker uses hexadecimal strings of 65 characters to uniquely identify
    containers, images and other objects. Docker's API almost always reports
    full IDs of 65 characters, but the ``docker`` program abbreviates these IDs
    to 12 characters in the user interface. We do the same because it makes the
    output more user friendly.

    :param id: A hexadecimal ID of 65 characters.
    :returns: A summarized ID of 12 characters.
    """
    return id[:12]

def slug(text):
    """
    Convert text to a "slug". Used by
    :py:attr:`redock.api.Container.ssh_alias`.

    :param text: The original text, e.g. "Some Random Text!".
    :returns: The slug text, e.g. "some-random-text".
    """
    slug = re.sub('[^a-z0-9]+', '-', text.lower())
    return slug.strip('-')

def create_configuration_directory():
    """
    Make sure Redock's local configuration directory exists.
    """
    if not os.path.isdir(REDOCK_CONFIG_DIR):
        logger.info("Creating directory: %s", format_path(REDOCK_CONFIG_DIR))
        os.makedirs(REDOCK_CONFIG_DIR)

# vim: ts=4 sw=4 et



# Minimal configuration management specialized to Ubuntu (Debian).
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: August 6, 2013
# URL: https://github.com/xolox/python-redock

"""
Bootstrap is a minimal `configuration management`_ system. Right now it's just
a toy module that I may or may not use to extend Redock beyond the existing
``redock start``, ``redock commit`` and ``redock kill`` functionality and
commands. Here is the design rationale behind Bootstrap (in its current form):

**Specialized towards Debian**
  Bootstrap is specialized towards Debian Linux (and its derivatives) because I
  have several years of hands on experience with Debian and Ubuntu Linux and
  because Docker currently gravitates to Ubuntu Linux (although this will
  probably change over time).

**Based on SSH connections**
  SSH_ is used to connect to remote hosts because it's the lowest common
  denominator that works with Docker_, VirtualBox_, XenServer_ and physical
  servers while being secure and easy to use.

**Remote code execution using Python**
  The execnet_ package is used to execute Python code on remote systems because
  I prefer the structure of Python code over shell scripts (Python avoids
  `quoting hell`_). When Bootstrap connects to a remote system it automatically
  installs the system package ``python2.7`` on the remote system because this
  is required to run execnet_ (on the other hand, execnet_ itself does not have
  to be installed on the remote system).

.. _configuration management: http://en.wikipedia.org/wiki/Configuration_management#Operating_System_configuration_management
.. _Docker: http://www.docker.io/
.. _execnet: http://codespeak.net/execnet/
.. _quoting hell: http://wiki.tcl.tk/1726
.. _SSH: http://en.wikipedia.org/wiki/Secure_Shell
.. _VirtualBox: https://www.virtualbox.org/
.. _XenServer: http://www.xenserver.org/
"""

# Standard library modules.
import os
import os.path
import pipes
import subprocess

# External dependencies.
from execnet import makegateway
from humanfriendly import Timer

# Modules included in our package.
from redock.logger import get_logger
from redock.utils import quote_command_line

MIRROR_FILE = os.path.expanduser('~/.redock/ubuntu-mirror.txt')

logger = get_logger(__name__)

class Bootstrap(object):

    """
    The Bootstrap configuration management system is implemented as the class
    :py:class:`Bootstrap`.
    """

    def __init__(self, ssh_alias):
        """
        Initialize the configuration management system by creating an execnet_
        gateway over an SSH connection. First we make sure the ``python2.7``
        package is installed; without it execnet_ won't work.

        :param ssh_alias: Alias of remote host in SSH client configuration.
        """
        self.logger = logger
        self.ssh_alias = ssh_alias
        # TODO Weaken requirement to just having "some version" of Python installed?
        self.logger.info("%s: Making sure the `python2.7' package is installed ..", self.ssh_alias)
        self.install_packages('python2.7')
        self.logger.info("%s: Initializing execnet over SSH connection ..", self.ssh_alias)
        # TODO Support sudo using makegateway('ssh=%s//python=sudo python')
        self.gateway = makegateway("ssh=%s" % self.ssh_alias)

    def upload_file(self, pathname, contents):
        """
        Create a file on the remote file system.

        :param pathname: The absolute pathname on the remote system.
        :param contents: The contents of the file (a string).
        """
        def remote_function(channel, pathname, contents):
            """
            Pure function that's executed remotely to create/update the file.
            """
            import os.path
            directory = os.path.dirname(pathname)
            if not os.path.isdir(directory):
                os.makedirs(directory)
            handle = open(pathname, 'w')
            handle.write(contents)
            bytes_written = handle.tell()
            handle.close()
            channel.send(bytes_written)
        channel = self.gateway.remote_exec(remote_function, pathname=pathname, contents=contents)
        bytes_written = channel.receive()
        if len(contents) != bytes_written:
            msg = "Remote side reported %i bytes written, but local side expected %i bytes!"
            raise Exception, msg % (bytes_written, len(contents))

    def install_packages(self, *packages):
        """
        Install the given system packages on the remote system.

        :param packages: The names of one or more packages to install (strings).
        """
        self.execute('apt-get', 'install', '-q', '-y', *packages)

    def update_system_packages(self):
        """
        Perform a full upgrade of all system packages on the remote system.
        """
        self.execute('apt-get', 'dist-upgrade', '-q', '-y', '--no-install-recommends')

    def execute(self, *command, **kw):
        """
        Execute a remote command over SSH so that the output of the remote
        command (the standard output and standard error streams) is immediately
        visible on the local terminal. If no standard input is given, this
        allocates a pseudo-tty_ (using ``ssh -t``) which means the operator can
        interact with the remote system should it prompt for input.

        Raises :py:exc:`ExternalCommandFailed` if the remote command ends with a
        nonzero exit code.

        :param command: A list with the remote command and its arguments.
        :param input: The standard input for the command (expected to be a
                      string). This is an optional keyword argument. If this
                      argument is given, no pseudo-tty_ will be allocated.

        .. _pseudo-tty: http://en.wikipedia.org/wiki/Pseudo_terminal
        """
        has_input = kw.get('input') is not None
        ssh_command = ['ssh']
        if not has_input:
            ssh_command.append('-t')
        ssh_command.append(self.ssh_alias)
        ssh_command.extend(command)
        self.logger.info("%s: Executing command %s", self.ssh_alias, ' '.join(ssh_command))
        options = dict()
        if has_input:
            options['stdin'] = subprocess.PIPE
        process = subprocess.Popen(ssh_command, **options)
        process.communicate(kw.get('input'))
        self.logger.debug("%s: Command exited with status %i.", self.ssh_alias, process.returncode)
        if process.returncode != 0:
            msg = "Remote command on %s failed with exit status %i! (command: %s)"
            raise ExternalCommandFailed, msg % (self.ssh_alias, process.returncode, ' '.join(command))

    def rsync(self, local_directory, remote_directory, cvs_exclude=True, delete=True):
        """
        Copy a directory on the host to the container using rsync over SSH.

        Raises :py:exc:`ExternalCommandFailed` if the remote command ends with a
        nonzero exit code.

        :param local_directory: The pathname of the source directory on the host.
        :param remote_directory: The pathname of the target directory in the container.
        :param cvs_exclude: Exclude version control files (enabled by default).
        :param delete: Delete remote files that don't exist locally (enabled by default).
        """
        rsync_timer = Timer()
        self.install_packages('rsync')
        def normalize(directory):
            """ Make sure a directory path ends with a trailing slash. """
            return "%s/" % directory.rstrip('/')
        location = "%s:%s" % (self.ssh_alias, normalize(remote_directory))
        self.logger.debug("Uploading %s to %s ..", local_directory, location)
        command = ['rsync', '-a']
        command.extend(['--rsync-path', 'mkdir -p %s && rsync' % pipes.quote(remote_directory)])
        if cvs_exclude:
            command.append('--cvs-exclude')
            command.extend(['--exclude', '.hgignore'])
        if delete:
            command.append('--delete')
        command.append(normalize(local_directory))
        command.append(location)
        self.logger.debug("Generated rsync command: %s", quote_command_line(command))
        exit_code = os.spawnvp(os.P_WAIT, command[0], command)
        if exit_code == 0:
            self.logger.debug("Finished upload using rsync in %s.", rsync_timer)
        self.logger.debug("rsync exited with status %d.", exit_code)
        if exit_code != 0:
            msg = "Failed to upload directory %s to %s, rsync exited with nonzero status %d! (command: %s)"
            raise ExternalCommandFailed, msg % (local_directory, location, exit_code, quote_command_line(command))

class ExternalCommandFailed(Exception):
    """
    Raised by :py:func:`Bootstrap.execute()` and :py:func:`Bootstrap.rsync()`
    when an external command fails (ends with a nonzero exit status).
    """

# vim: ts=4 sw=4 et



# Command line interface for Redock, a human friendly wrapper around Docker.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: August 6, 2013
# URL: https://github.com/xolox/python-redock

# Standard library modules.
import getopt
import logging
import os
import subprocess
import sys
import textwrap

# External dependencies.
import coloredlogs
from humanfriendly import Timer

# Modules included in our package.
from redock.api import Container, Image
from redock.logger import get_logger, root_logger

logger = get_logger(__name__)

def main():
    """
    Command line interface for the ``redock`` program.
    """
    # Initialize coloredlogs.
    coloredlogs.install()
    # Parse and validate the command line arguments.
    try:
        # Command line option defaults.
        hostname = None
        message = None
        # Parse the command line options.
        options, arguments = getopt.getopt(sys.argv[1:], 'b:n:m:vh',
                                          ['hostname=', 'message=', 'verbose', 'help'])
        for option, value in options:
            if option in ('-n', '--hostname'):
                hostname = value
            elif option in ('-m', '--message'):
                message = value
            elif option in ('-v', '--verbose'):
                if root_logger.getEffectiveLevel() == logging.INFO:
                    root_logger.setLevel(logging.VERBOSE)
                elif root_logger.getEffectiveLevel() == logging.VERBOSE:
                    root_logger.setLevel(logging.DEBUG)
            elif option in ('-h', '--help'):
                usage()
                return
            else:
                # Programming error...
                assert False, "Unhandled option!"
        # Handle the positional arguments.
        if len(arguments) < 2:
            usage()
            return
        supported_actions = ('start', 'commit', 'kill', 'delete')
        action = arguments.pop(0)
        if action not in supported_actions:
            msg = "Action not supported: %r (supported actions are: %s)"
            raise Exception, msg % (action, ', '.join(supported_actions))
    except Exception, e:
        logger.error("Failed to parse command line arguments!")
        logger.exception(e)
        usage()
        sys.exit(1)
    # Start the container and connect to it over SSH.
    try:
        for image_name in arguments:
            container = Container(image=Image.coerce(image_name),
                                  hostname=hostname)
            if action == 'start':
                container.start()
                if len(arguments) == 1 and all(os.isatty(n) for n in range(3)):
                    ssh_timer = Timer()
                    logger.info("Detected interactive terminal, connecting to container ..")
                    ssh_client = subprocess.Popen(['ssh', container.ssh_alias])
                    ssh_client.wait()
                    if ssh_client.returncode == 0:
                        logger.info("SSH client exited after %s.", ssh_timer)
                    else:
                        logger.warn("SSH client exited with status %i after %s.",
                                    ssh_client.returncode, ssh_timer)
            elif action == 'commit':
                container.commit(message=message)
            elif action == 'kill':
                container.kill()
            elif action == 'delete':
                container.delete()
            else:
                # Programming error...
                assert False, "Unhandled action!"
    except Exception, e:
        logger.exception(e)
        sys.exit(1)

def usage():
    """
    Print a usage message to the console.
    """
    print textwrap.dedent("""
        Usage: redock [OPTIONS] ACTION CONTAINER..

        Create and manage Docker containers and images. Supported actions are
        `start', `commit', `kill' and `delete'.

        Supported options:

          -n, --hostname=NAME  set container host name (defaults to image tag)
          -m, --message=TEXT   message for image created with `commit' action
          -v, --verbose        make more noise (can be repeated)
          -h, --help           show this message and exit
    """).strip()

# vim: ts=4 sw=4 et



# Initialization of the base image used by Redock.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 15, 2013
# URL: https://github.com/xolox/python-redock

"""
The :py:mod:`redock.base` module implements the initialization of the base
image used by Redock. You'll probably never need to use this module directly
because :py:func:`redock.api.Container.start()` calls
:py:func:`find_base_image()` and :py:func:`create_base_image()` as needed.
"""

# Standard library modules.
import pipes

# External dependencies.
from humanfriendly import Timer

# Modules included in our package.
from redock.logger import get_logger
from redock.utils import (RemoteTerminal, get_ssh_public_key,
                          select_ubuntu_mirror, summarize_id)

# The logger for this module.
logger = get_logger(__name__)

# The repository and tag of Redock's base image.
BASE_IMAGE_REPO = 'redock'
BASE_IMAGE_TAG = 'base'
BASE_IMAGE_NAME = '%s:%s' % (BASE_IMAGE_REPO, BASE_IMAGE_TAG)
SSHD_LOG_FILE = '/var/log/sshd.log'

APT_CONFIG = '''
# /etc/apt/apt.conf.d/90redock:
# Disable automatic installation of recommended packages. Debian doesn't do
# this; the policy change came from Ubuntu, and I don't like it one bit!
# Fortunately we can still change it :-)

APT::Install-Recommends "false";

# vim: ft=perl
'''

SOURCES_LIST = '''
# /etc/apt/sources.list: Use a local package mirror.

deb {mirror} precise main universe

# vim: ft=debsources
'''

SUPERVISOR_CONFIG = '''
# /etc/supervisor/conf.d/ssh-server.conf:
# Replacement for /etc/init/ssh.conf that doesn't need upstart.

[program:ssh-server]
command = bash -c 'mkdir -p -m0755 /var/run/sshd && /usr/sbin/sshd -eD'
# stdout_logfile = /var/log/supervisor/ssh-server.log
# redirect_stderr = true
autorestart = true

# vim: ft=dosini
'''.format(log_file=SSHD_LOG_FILE)

def find_base_image(client):
    """
    Find the id of the base image that's used by Redock to create new
    containers. If the image doesn't exist yet it will be created using
    :py:func:`create_base_image()`.

    :param client: Connection to Docker (instance of :py:class:`docker.Client`)
    :returns: The unique id of the base image.
    """
    logger.verbose("Looking for base image ..")
    image_id = find_named_image(client, BASE_IMAGE_REPO, BASE_IMAGE_TAG)
    if image_id:
        logger.verbose("Found base image: %s", summarize_id(image_id))
        return image_id
    else:
        logger.verbose("No base image found, creating it ..")
        return create_base_image(client)

def create_base_image(client):
    """
    Create the base image that's used by Redock to create new containers. This
    base image differs from the ubuntu:precise_ image (on which it is based) on
    a couple of points:

    - Automatic installation of recommended packages is disabled to conserve
      disk space.

    - The Ubuntu package mirror is set to a geographically close location to
      speed up downloading of system packages (see
      :py:func:`redock.utils.select_ubuntu_mirror()`).

    - The package list is updated to make sure apt-get_ installs the most up to
      date packages.

    - The following system packages are installed:

      language-pack-en-base_
        In a base Docker Ubuntu 12.04 image lots of commands complain loudly
        about the locale_. This silences the warnings by fixing the problem
        (if you want to call it that).

      openssh-server_
        After creating a new container Redock will connect to it over SSH_,
        so having an SSH server installed is a good start :-)

      supervisor_
        The base Docker Ubuntu 12.04 image has init_ (upstart_) disabled.
        Indeed we don't need all of the bagage that comes with init but it is
        nice to have a process runner for the SSH_ server (and eventually maybe
        more).

    - The initscripts_ and upstart_ system packages are marked 'on hold' so
      that apt-get_ will not upgrade them. This makes it possible to run
      ``apt-get dist-upgrade`` inside containers.

    - An SSH_ key pair is generated and the SSH public key is installed inside
      the base image so that Redock can connect to the container over SSH (you
      need ssh-keygen_ installed).

    - Supervisor_ is configured to automatically start the SSH_ server.

    :param client: Connection to Docker (instance of :py:class:`docker.Client`)
    :returns: The unique id of the base image.

    .. _apt-get: http://manpages.ubuntu.com/manpages/precise/man8/apt-get.8.html
    .. _init: http://manpages.ubuntu.com/manpages/precise/man8/init.8.html
    .. _initscripts: http://packages.ubuntu.com/precise/initscripts
    .. _language-pack-en-base: http://packages.ubuntu.com/precise/language-pack-en-base
    .. _locale: http://en.wikipedia.org/wiki/Locale
    .. _openssh-server: http://packages.ubuntu.com/precise/openssh-server
    .. _ssh-keygen: http://manpages.ubuntu.com/manpages/precise/man1/ssh-keygen.1.html
    .. _supervisor: http://packages.ubuntu.com/precise/supervisor
    .. _ubuntu:precise: https://index.docker.io/_/ubuntu/
    .. _upstart: http://packages.ubuntu.com/precise/upstart
    """
    download_image(client, 'ubuntu', 'precise')
    creation_timer = Timer()
    logger.info("Initializing base image (this can take a few minutes but you only have to do it once) ..")
    command = ' && '.join([
        'echo %s > /etc/apt/apt.conf.d/90redock' % pipes.quote(APT_CONFIG.strip()),
        'echo %s > /etc/apt/sources.list' % pipes.quote(SOURCES_LIST.format(mirror=select_ubuntu_mirror()).strip()),
        'apt-get update',
        'DEBIAN_FRONTEND=noninteractive apt-get install -q -y language-pack-en-base openssh-server supervisor',
        'apt-get clean', # Don't keep the +/- 20 MB of *.deb archives after installation.
        # Make it possible to run `apt-get dist-upgrade'.
        # https://help.ubuntu.com/community/PinningHowto#Introduction_to_Holding_Packages
        'apt-mark hold initscripts upstart',
        # Install the generated SSH public key.
        'mkdir -p /root/.ssh',
        'echo %s > /root/.ssh/authorized_keys' % pipes.quote(get_ssh_public_key()),
        # Create the Supervisor configuration for the SSH server.
        'echo %s > /etc/supervisor/conf.d/ssh-server.conf' % pipes.quote(SUPERVISOR_CONFIG.strip())])
    logger.debug("Generated command line: %s", command)
    result = client.create_container(image='ubuntu:precise',
                                     command='bash -c %s' % pipes.quote(command),
                                     hostname='redock-template',
                                     ports=['22'])
    container_id = result['Id']
    for text in result.get('Warnings', []):
      logger.warn("%s", text)
    logger.verbose("Created container %s.", summarize_id(container_id))
    client.start(container_id)
    with RemoteTerminal(container_id):
        logger.info("Waiting for initialization to finish ..")
        client.wait(container_id)
        logger.info("Finished initialization in %s.", creation_timer)
    commit_timer = Timer()
    logger.info("Saving initialized container as new base image ..")
    result = client.commit(container_id, repository='redock', tag='base')
    logger.info("Done! Committed base image as %s in %s.", summarize_id(result['Id']), commit_timer)
    return result['Id']

def find_named_image(client, repository, tag):
    """
    Find the most recent Docker image with the given repository and tag.

    :param repository: The name of the image's repository.
    :param tag: The name of the image's tag.
    :returns: The unique id of the most recent image available, or ``None`` if
              no images were matched.
    """
    matches = []
    for image in client.images():
        if image.get('Repository') == repository and image.get('Tag') == tag:
            matches.append(image)
    if matches:
        matches.sort(key=lambda i: i['Created'])
        return matches[-1]['Id']

def download_image(client, repository, tag):
    """
    Download the requested image. If the image is already available locally it
    won't be downloaded again.

    :param client: Connection to Docker (instance of :py:class:`docker.Client`)
    :param repository: The name of the image's repository.
    :param tag: The name of the image's tag.
    """
    if not find_named_image(client, repository, tag):
        download_timer = Timer()
        logger.info("Downloading image %s:%s (please be patient, this can take a while) ..", repository, tag)
        client.pull(repository=repository, tag=tag)
        logger.info("Finished downloading image in %s.", download_timer)

# vim: ts=4 sw=4 et



# Main API for Redock, a human friendly wrapper around Docker.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2013
# URL: https://github.com/xolox/python-redock

"""
The :py:mod:`redock.api` module defines two classes and two exception types:

- :py:class:`Container`
- :py:class:`Image`
- :py:class:`NoContainerRunning`
- :py:class:`SecureShellTimeout`
"""

# Standard library modules.
import os
import pipes
import socket
import subprocess
import sys
import textwrap
import time

# External dependencies.
import docker
import humanfriendly
import update_dotdee

# Initialize the logger.
from redock.logger import get_logger
from redock.base import BASE_IMAGE_NAME, find_base_image
from redock.utils import (PRIVATE_SSH_KEY, Config, RemoteTerminal,
                          find_local_ip_addresses, quote_command_line,
                          slug, summarize_id)

logger = get_logger(__name__)

class Container(object):

    """
    :py:class:`Container` is the main entry point to the Redock API. It aims to
    provide a simple to use representation of Docker containers (and in
    extension Docker images). You'll probably never need most of the methods
    defined in this class; if you're getting started with Redock you should
    focus on these methods:

    - :py:func:`Container.start()`
    - :py:func:`Container.commit()`
    - :py:func:`Container.kill()`

    After you create and start a container with Redock you can do with the
    container what you want by starting out with an SSH_ connection. When
    you're done you either save your changes or discard them and kill the
    container. That's probably all you need from Redock :-)

    .. _SSH: http://en.wikipedia.org/wiki/Secure_Shell
    """

    def __init__(self, image, hostname=None, timeout=10):
        """
        Initialize a :py:class:`Container` from the given arguments.

        :param image: The repository and tag of the container's image (in the
                      format expected by :py:class:`Image.coerce()`).
        :param hostname: The host name to use inside the container. If none is
                         given, the image's tag is used.
        :param timeout: The timeout in seconds while waiting for a container to
                        become reachable over SSH_ (a couple of seconds should
                        be plenty).
        """
        # Validate and store the arguments.
        self.image = Image.coerce(image)
        self.base = Image.coerce(BASE_IMAGE_NAME)
        self.hostname = hostname or self.image.tag
        self.timeout = timeout
        # Initialize some private variables.
        self.logger = logger
        self.config = Config()
        self.session = Session()
        self.update_dotdee = update_dotdee.UpdateDotDee(os.path.expanduser('~/.ssh/config'))
        # Connect to the Docker API over HTTP.
        try:
            self.logger.debug("Connecting to Docker daemon ..")
            self.client = docker.Client()
            self.logger.debug("Successfully connected to Docker.")
        except Exception, e:
            self.logger.error("Failed to connect to Docker!")
            self.logger.exception(e)
            raise

    def start(self):
        """
        Create and start the Docker container. On the first run of Redock this
        creates a base image using :py:func:`redock.base.create_base_image()`.
        """
        if not self.find_container():
            if not self.find_image(self.image):
                self.logger.info("Image doesn't exist yet, creating it: %r", self.image)
                self.base.id = find_base_image(self.client)
            self.start_supervisor()
        self.setup_ssh_access()

    def commit(self, message=None, author=None):
        """
        Commit any changes to the running container to the associated image.
        Corresponds to the ``docker commit`` command.

        Raises :py:exc:`NoContainerRunning` if an associated Docker container
        is not already running.

        :param message: A short message describing the commit (a string).
        :param author: The name of the author (a string).
        """
        self.check_active()
        self.logger.info("Committing changes: %s", message or 'no description given')
        result = self.client.commit(self.session.container_id, repository=self.image.repository,
                                    tag=self.image.tag, message=message, author=author)
        image_ids = [i['Id'] for i in self.client.images()]
        self.image.id = self.expand_id(result['Id'], image_ids)

    def kill(self):
        """
        Kill and remove the container. All changes since the last time that
        :py:func:`Container.commit()` was called will be lost.
        """
        if self.find_container():
            if self.session.remote_terminal:
                self.session.remote_terminal.detach()
            self.logger.info("Killing container ..")
            self.client.kill(self.session.container_id)
            self.logger.info("Removing container ..")
            self.client.remove_container(self.session.container_id)
            with self.config as state:
                del state['containers'][self.image.key]
            self.session.reset()
        self.revoke_ssh_access()

    def delete(self):
        """
        Delete the image associated with the container (if any). The data in
        the image will be lost.
        """
        self.logger.info("Deleting image %s ..", self.image.name)
        self.client.remove_image(self.image.name)

    def find_container(self):
        """
        Check to see if the current :py:class:`Container` has an associated
        Docker container that is currently running.

        :returns: ``True`` when a running container exists, ``False``
                  otherwise.
        """
        if not self.session.container_id:
            self.logger.verbose("Looking for running container ..")
            state = self.config.load()
            container_id = state['containers'].get(self.image.key)
            # Make sure the container is still running.
            if container_id in [c['Id'] for c in self.client.containers()]:
                self.session.container_id = container_id
                self.logger.info("Found running container: %s", summarize_id(container_id))
        return bool(self.session.container_id)

    def find_image(self, image_to_find):
        """
        Find the most recent Docker image with the given repository and tag.

        :param image_to_find: The :py:class:`Image` we're looking for.
        :returns: The most recent :py:class:`Image` available, or ``None`` if
                  no images were matched.
        """
        matches = []
        for image in self.client.images():
            if (image.get('Repository') == image_to_find.repository
                    and image.get('Tag') == image_to_find.tag):
                matches.append(image)
        if matches:
            matches.sort(key=lambda i: i['Created'])
            image = matches[-1]
            return Image(repository=image['Repository'],
                         tag=image['Tag'],
                         id=image['Id'])

    def start_supervisor(self):
        """
        Starts the container and runs Supervisor inside the container.
        """
        command = '/usr/bin/supervisord -n'
        self.logger.info("Starting process supervisor (and SSH server) ..")
        # Select the Docker image to use as a base for the container.
        image = self.find_image(self.image) or self.find_image(self.base)
        self.logger.verbose("Creating container from image: %r", image)
        # Start the container with the given command.
        result = self.client.create_container(image=image.unique_name,
                                              command=command,
                                              hostname=self.hostname,
                                              ports=['22'])
        container_ids = [c['Id'] for c in self.client.containers(all=True)]
        self.session.container_id = self.expand_id(result['Id'], container_ids)
        self.logger.verbose("Created container: %s", summarize_id(self.session.container_id))
        for text in result.get('Warnings', []):
            logger.warn("%s", text)
        # Start the command inside the container.
        self.logger.verbose("Running command: %s", command)
        self.client.start(self.session.container_id)
        # Make the output from the container visible to the user.
        self.session.remote_terminal = RemoteTerminal(self.session.container_id)
        self.session.remote_terminal.attach()
        # Persist association between (repository, tag) and container id.
        with self.config as state:
            state['containers'][self.image.key] = self.session.container_id

    def get_ssh_client_command(self, ip_address=None, port_number=None):
        """
        Generate an SSH_ client command line that connects to the container
        (assumed to be running).

        :param ip_address: This optional argument overrides the default IP
                           address (which is otherwise automatically
                           discovered).
        :param port_number: This optional argument overrides the default port
                            number (which is otherwise automatically
                            discovered).
        :returns: The SSH client command line as a list of strings containing
                  the command and its arguments.
        """
        command = ['ssh']
        # Connect as the root user inside the container.
        command.extend(['-l', 'root'])
        # Connect using the generated SSH private key.
        command.extend(['-i', PRIVATE_SSH_KEY])
        # Don't check or store the host key (it's pointless).
        command.extend(['-o', 'StrictHostKeyChecking=no'])
        command.extend(['-o', 'UserKnownHostsFile=/dev/null'])
        # Silence the message "Warning: Permanently added ... to the list of known hosts."
        command.append('-q')
        # Connect through a NAT port on the local system.
        if not (ip_address and port_number):
            ip_address, port_number = self.ssh_endpoint
        command.extend(['-p', str(port_number)])
        # Make the SSH connection binary safe.
        command.extend(['-e', 'none'])
        # Finish the command by including the IP address.
        command.append(ip_address)
        self.logger.debug("Generated SSH command: %s", quote_command_line(command))
        # Return the generated command.
        return command

    @property
    def ssh_alias(self):
        """
        Get the SSH_ alias that should be used to connect to the container.
        """
        return slug(self.hostname + '-container')

    def setup_ssh_access(self):
        """
        Update ``~/.ssh/config`` to make it easy to connect to the container
        over SSH_ from the host system. This generates a host definition to
        include in the SSH client configuration file and uses update-dotdee_ to
        merge the generated host definition with the user's existing SSH client
        configuration file.

        .. _update-dotdee: https://pypi.python.org/pypi/update-dotdee
        """
        self.logger.verbose("Configuring SSH access ..")
        self.update_dotdee.create_directory()
        with open(self.ssh_config_file, 'w') as handle:
            handle.write(textwrap.dedent("""
                Host {alias}
                  Hostname {address}
                  Port {port}
                  User root
                  IdentityFile {key}
                  StrictHostKeyChecking no
                  UserKnownHostsFile /dev/null
            """.format(alias=self.ssh_alias,
                       address=self.ssh_endpoint[0],
                       port=self.ssh_endpoint[1],
                       key=PRIVATE_SSH_KEY,
                       redock=pipes.quote(os.path.abspath(sys.argv[0])),
                       container=pipes.quote(self.image.name))))
        self.update_dotdee.update_file()
        self.logger.info("Successfully configured SSH access. Use this command: ssh %s", self.ssh_alias)

    def revoke_ssh_access(self):
        """
        Remove the container's SSH_ client configuration from
        ``~/.ssh/config``.
        """
        self.logger.info("Removing SSH client configuration ..")
        if os.path.isfile(self.ssh_config_file):
            os.unlink(self.ssh_config_file)
        self.update_dotdee.update_file()

    @property
    def ssh_config_file(self):
        """
        Get the pathname of the SSH_ client configuration for the container.
        """
        return os.path.expanduser('~/.ssh/config.d/redock:%s' % self.image.name)

    @property
    def ssh_endpoint(self):
        """
        Wait for the container to become reachable over SSH_ and get a tuple
        with the IP address and port number that can be used to connect to the
        container over SSH.
        """
        self.check_active()
        if self.session.ssh_endpoint:
            return self.session.ssh_endpoint
        # Get the local port connected to the container.
        host_port = int(self.client.port(self.session.container_id, '22'))
        self.logger.debug("Configured port redirection for container %s: %s:%i -> %s:%i",
                          summarize_id(self.session.container_id),
                          socket.gethostname(), host_port,
                          self.hostname, 22)
        # Give the container time to finish the SSH server installation.
        self.logger.verbose("Waiting for SSH connection to %s (max %i seconds) ..",
                            summarize_id(self.session.container_id), self.timeout)
        global_timeout = time.time() + self.timeout
        ssh_timer = humanfriendly.Timer()
        while time.time() < global_timeout:
            for ip_address in find_local_ip_addresses():
                # Try to open an SSH connection to the container.
                self.logger.debug("Connecting to container over SSH at %s:%s ..", ip_address, host_port)
                command = self.get_ssh_client_command(ip_address, host_port) + ['true']
                ssh_client = subprocess.Popen(command, stdin=open(os.devnull), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                # Give this attempt at most 10 seconds to succeed.
                inner_timeout = time.time() + 10
                while time.time() < inner_timeout:
                    if ssh_client.poll() is not None:
                        break
                    time.sleep(0.1)
                else:
                    self.logger.debug("Attempt to connect timed out!")
                if ssh_client.returncode == 0:
                    # At this point we have successfully connected!
                    self.session.ssh_endpoint = (ip_address, host_port)
                    self.logger.debug("Connected to %s at %s using SSH in %s.",
                                      self.image.name, self.session.ssh_endpoint,
                                      ssh_timer)
                    return self.session.ssh_endpoint
                time.sleep(1)
        msg = "Time ran out while waiting to connect to container %s over SSH! (Most likely something went wrong while initializing the container..)"
        raise SecureShellTimeout, msg % self.image.name

    def __repr__(self):
        """
        Pretty print a :py:class:`Container` object.
        """
        template = "Container(image=%r, base=%r, hostname=%r)"
        return template % (self.image, self.base, self.hostname)

    # Miscellaneous methods.

    def check_active(self):
        """
        Check if the :py:class:`Container` is associated with a running Docker
        container. If no running Docker container is found,
        :py:class:`NoContainerRunning` is raised.
        """
        if not self.find_container():
            raise NoContainerRunning, "No active container!"

    def expand_id(self, short_id, candidate_ids):
        """
        :py:func:`docker.Client.create_container()` and
        :py:func:`docker.Client.commit()` report short ids (12 characters)
        while :py:func:`docker.Client.containers()` and
        :py:func:`docker.Client.images()` report long ids (65 characters). I'd
        rather use the full ids where possible. This method translates short
        ids into long ids at the expense of an additional API call (who
        cares).

        Raises :py:exc:`exceptions.Exception` if no long id corresponding to
        the short id can be matched (this might well be a purely theoretical
        problem, it certainly shouldn't happen during regular use).

        :param short_id: A short id of 12 characters.
        :param candidate_ids: A list of available long ids.
        :returns: The long id corresponding to the given short id.
        """
        self.logger.debug("Translation short id %s into long id ..", short_id)
        for long_id in candidate_ids:
            self.logger.debug("Checking candidate: %s", long_id)
            if long_id.startswith(short_id):
                return long_id
        msg = "Failed to translate short id (%s) into long id!"
        raise Exception, msg % short_id

class Image(object):

    """
    Simple representation of Docker images.
    """

    def __init__(self, repository, tag, id=None):
        """
        Initialize an :py:class:`Image` instance from the given arguments.

        :param repository: The name of the image's repository.
        :param tag: The image's tag (name).
        :param id: The unique hash of the image (optional).
        """
        self.repository = repository
        self.tag = tag
        self.id = id

    @staticmethod
    def coerce(value):
        """
        Coerce strings to :py:class:`Image` objects.

        Raises :py:exc:`exceptions.ValueError` when a string with an
        incorrect format is given.

        :param value: The name of the image, expected to be a string of the
                      form ``repository:tag``. If an :py:class:`Image` object
                      is given it is returned unmodified.
        :returns: An :py:class:`Image` object.
        """
        if isinstance(value, basestring):
            components = value.split(':')
            if len(components) == 1:
                value = Image(repository=os.environ['USER'],
                              tag=components[0])
            else:
                if len(components) != 2:
                    msg = "Invalid image name (expected 'repository:tag', got %r)"
                    raise ValueError, msg % value
                value = Image(repository=components[0],
                              tag=components[1])
        return value

    @property
    def key(self):
        """
        Get a tuple with the image's repository and tag.
        """
        return (self.repository, self.tag)

    @property
    def name(self):
        """
        Get the human readable name of an :py:class:`Image` as a string of the
        form ``repository:tag``.
        """
        return "%s:%s" % (self.repository, self.tag)

    @property
    def unique_name(self):
        """
        Get the machine readable unique name of an :py:class:`Image`. If the
        image has a unique hash that will be used, otherwise a string of the
        form ``repository:tag`` is returned.
        """
        if self.id:
            return self.id
        else:
            return self.name

    def __repr__(self):
        """
        Provide a textual representation of an :py:class:`Image` object.
        """
        properties = ["repository=%r" % self.repository,
                      "tag=%r" % self.tag]
        if self.id:
            properties.append("id=%r" % summarize_id(self.id))
        return "Image(%s)" % ", ".join(properties)

class Session(object):

    """
    Dumb object to hold session variables associated with a running Docker
    container.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        """
        Reset all known session variables to ``None``.
        """
        self.container_id = None
        self.custom_image = None
        self.remote_terminal = None
        self.ssh_endpoint = None

class SecureShellTimeout(Exception):
    """
    Raised by :py:attr:`Container.ssh_endpoint` when Redock fails to connect to
    the Docker container within a reasonable amount of time (10 seconds by
    default).
    """

class NoContainerRunning(Exception):
    """
    Raised by :py:func:`Container.check_active` when a :py:class:`Container`
    doesn't have an associated Docker container running.
    """

# vim: ts=4 sw=4 et



# Logging configuration for Redock.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: August 6, 2013
# URL: https://github.com/xolox/python-redock

# Standard library modules.
import logging
import os

# External dependencies.
import coloredlogs
import verboselogs

# Configure the root logger.
root_logger = logging.getLogger()
if 'REDOCK_DEBUG' in os.environ:
    root_logger.setLevel(logging.DEBUG)
else:
    root_logger.setLevel(logging.INFO)

# Silence the logger of an external dependency.
logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARN)

def get_logger(name):
    return verboselogs.VerboseLogger(name)

# vim: ts=4 sw=4 et



#!/usr/bin/env python

# Setup script for redock.
#
# Author: Peter Odding <peter@peterodding.com>
# Last Change: July 16, 2013
# URL: https://github.com/xolox/python-redock

import re
from os.path import abspath, dirname, join
from setuptools import setup, find_packages

# Find the directory containing the source distribution.
source_directory = dirname(abspath(__file__))

# Find the current version.
module = join(source_directory, 'redock', '__init__.py')
for line in open(module, 'r'):
    match = re.match(r'^__version__\s*=\s*["\']([^"\']+)["\']$', line)
    if match:
        version_string = match.group(1)
        break
else:
    raise Exception, "Failed to extract version from redock/__init__.py!"

# Fill in the long description (for the benefit of PyPi)
# with the contents of README.rst (rendered by GitHub).
readme_text = open(join(source_directory, 'README.rst'), 'r').read()

# Fill in the "install_requires" field based on requirements.txt.
requirements = [l.strip() for l in open(join(source_directory, 'requirements.txt'), 'r')]

setup(name='redock',
      version=version_string,
      description="Human friendly wrapper around Docker",
      long_description=readme_text,
      url='https://github.com/xolox/python-redock',
      author='Peter Odding',
      author_email='peter@peterodding.com',
      packages=find_packages(),
      entry_points=dict(console_scripts=['redock = redock.cli:main']),
      install_requires=requirements,
      test_suite='redock.tests')
